import logging
from argparse import Namespace
from typing import Callable

from fastapi import FastAPI

from talisman_tools.commands.servers.commands.processor.methods.process import register_process_docs
from talisman_tools.commands.servers.commands.processor.methods.update import register_update
from talisman_tools.commands.servers.server_helper import async_register_context_manager, register_exception_handlers
from talisman_tools.plugin import EndpointPlugins
from tp_interfaces.abstract import AbstractDocumentProcessor, AbstractUpdatableModel
from tp_interfaces.domain.manager import DomainManager

logger = logging.getLogger(__name__)


def register_process(app: FastAPI, processor):
    async_register_context_manager(app, processor)

    @app.on_event("startup")
    def register():
        register_process_docs(
            app=app,
            endpoint='/',
            processor=processor,
            logger=logger
        )


def action(args: Namespace, processor_factory: Callable[[Namespace], AbstractDocumentProcessor]) -> FastAPI:
    processor = processor_factory(args)
    logger.info(f"Loaded {processor}")

    app = FastAPI(title="talisman-ie REST server", description=f"talisman-ie REST server for {processor}")

    if isinstance(processor, AbstractUpdatableModel):
        register_update(app, processor, processor.update_type)

    for endpoint, register in EndpointPlugins.flattened.items():
        register()(app=app, endpoint=endpoint, processor=processor, logger=logger)

    register_exception_handlers(app, logger=logger)
    if DomainManager():
        async_register_context_manager(app, DomainManager())
    register_process(app, processor)

    return app



import asyncio
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Callable

from talisman_tools.arguments.reader import get_reader_factory
from talisman_tools.commands.model.commands.evaluate.action import mode  # TODO: pulling unused dependencies
from tp_interfaces.abstract import AbstractDocumentProcessor


def configure_evaluate_parser(parser: ArgumentParser) -> None:
    parser.add_argument('eval_mode', type=str, choices=set(mode), metavar='<evaluation mode>')
    parser.add_argument('--config', type=Path, metavar='<document processing config path>')
    parser.add_argument('--eval_config', type=Path, metavar='<evaluation config path>')

    reader_factory = get_reader_factory(parser, {'gold_path': '<path to gold>', '-input_docs': '<input documents path>'})

    def get_action() -> Callable[[AbstractDocumentProcessor, Namespace], None]:
        def action(processor: AbstractDocumentProcessor, args: Namespace) -> None:
            from .action import evaluate
            readers = reader_factory(args)
            asyncio.run(evaluate(
                processor, args.eval_mode, readers['gold_path'], args.config, args.eval_config, input_reader=readers.get('-input_docs')
            ))

        return action

    parser.set_defaults(processor_action=get_action)



from pathlib import Path
from typing import Callable, Dict

from tp_interfaces.readers.abstract import AbstractConfigurableReader, AbstractReader
from .abstract import AbstractPluginManager, flattened


class ReaderPluginManager(AbstractPluginManager):
    def __init__(self):
        AbstractPluginManager.__init__(self, 'READERS')

    @property
    def flattened(self) -> Dict[str, Callable[[Path], AbstractReader]]:
        return flattened(self.plugins)

    @staticmethod
    def _check_value(value) -> bool:
        return isinstance(value, dict)  # TODO: improve validation


class ConfigurableReaderPluginManager(AbstractPluginManager):
    def __init__(self):
        AbstractPluginManager.__init__(self, 'CONFIGURABLE_READERS')

    @property
    def flattened(self) -> Dict[str, Callable[[dict], AbstractConfigurableReader]]:
        return flattened(self.plugins)

    @staticmethod
    def _check_value(value) -> bool:
        return isinstance(value, dict)  # TODO: improve validation



from functools import partial

from .disambiguation_quality import evaluate_dmb
from .nerc import evaluate_nerc
from .relext import evaluate_relext, evaluate_relext_upper_bound


evaluators = {
    'all': {
        'nerc': evaluate_nerc,
        'relext': evaluate_relext,
        'relext-upper-bound': evaluate_relext_upper_bound,
        'dmb': partial(evaluate_dmb, at_ks=[1, 2, 3])  # TODO: make configurable from cli
    },
    'relext': {
        'relext': evaluate_relext
    },
    'nerc': {
        'nerc': evaluate_nerc
    },
    'dmb': {
        'dmb': partial(evaluate_dmb, at_ks=[1, 2, 3])  # TODO: make configurable from cli
    },
    # 'coref': {
    #     'coref': evaluate_coref
    # }
}



from typing import Callable, FrozenSet

from talisman_tools.configure.wrapper import DEFAULT_WRAPPER_ACTIONS
from tp_interfaces.abstract import AbstractDocumentProcessor
from .abstract import AbstractPluginManager


class WrapperActionsPluginManager(AbstractPluginManager):
    def __init__(self):
        AbstractPluginManager.__init__(self, 'WRAPPER_ACTIONS', DEFAULT_WRAPPER_ACTIONS)

    def add_wrapper_action(self, key: FrozenSet[str], action: Callable[[dict], AbstractDocumentProcessor]) -> None:
        # TODO: rewrite it with some context menager that restore initial PM state on exit
        self._registered_plugins[None][key] = action

    @staticmethod
    def _check_value(value) -> bool:
        return isinstance(value, dict) and all(WrapperActionsPluginManager._validate_key_value(k, v) for k, v in value.items())

    @staticmethod
    def _validate_key_value(key, value) -> bool:
        return isinstance(key, frozenset)



import logging
from argparse import ArgumentParser, Namespace
from typing import Callable

from jsonformatter import JsonFormatter


def create_json_formatter() -> JsonFormatter:
    return JsonFormatter(
        fmt={"level": "levelname", "msg": "message", "time": "asctime"},
        datefmt="%Y-%m-%dT%H:%M:%S%z",
        mix_extra=True,
        skipkeys=True,
        ensure_ascii=False,
        default=lambda o: "NON-SERIALIZABLE"
    )


def configure_json_logging(level=logging.INFO):
    stderr_handler = logging.StreamHandler()
    json_formatter = create_json_formatter()
    stderr_handler.setFormatter(json_formatter)
    logging.basicConfig(level=level, handlers=[stderr_handler])


def get_logging_factory(parser: ArgumentParser) -> Callable[[Namespace], None]:
    logging_levels = {
        'info': logging.INFO,
        'debug': logging.DEBUG,
        'warning': logging.WARNING,
        'error': logging.ERROR
    }
    parser.add_argument('--logging_level', choices=list(logging_levels.keys()), default='info')

    def configure_logging(args: Namespace) -> None:
        logging_level = logging_levels[args.logging_level]
        configure_json_logging(level=logging_level)

    return configure_logging



import logging
import time
from collections import defaultdict
from typing import Any, Callable

from tdm import TalismanDocument

from tp_interfaces.abstract import AbstractDocumentProcessor


logger = logging.getLogger(__name__)


def measure_default(f: Callable):
    def wrapper(*args, **kwargs) -> dict[str, float]:
        start_time = time.time()
        f(*args, **kwargs)
        end_time = time.time()

        return {'time': end_time - start_time}
    return wrapper


def measure_linux(f: Callable[[Any], dict[str, float]]):  # throws ImportError if resource is not available on the platform
    import resource

    def wrapper(*args, **kwargs) -> dict[str, float]:
        start_usage = resource.getrusage(resource.RUSAGE_SELF)
        res = f(*args, **kwargs)
        end_usage = resource.getrusage(resource.RUSAGE_SELF)

        res['ru_maxrss'] = end_usage.ru_maxrss  # there is no way to reset max_rss tracker
        res['ru_utime'] = end_usage.ru_utime - start_usage.ru_utime
        res['ru_stime'] = end_usage.ru_stime - start_usage.ru_stime
        return res
    return wrapper


def measure_torch(f: Callable[[Any], dict[str, float]]):  # throws ImportError if torch is not available
    import torch

    def wrapper(*args, **kwargs) -> dict[str, float]:
        res = f(*args, **kwargs)

        if not torch.cuda.is_available():
            return res

        for device in range(torch.cuda.device_count()):
            torch.cuda.init()  # for some reason without it stats is an empty dict
            stats = torch.cuda.memory_stats()

            res[f'device.{device}.peak_reserved'] = stats['reserved_bytes.all.peak']
            res[f'device.{device}.peak_allocated'] = stats['allocated_bytes.all.peak']
            torch.cuda.reset_peak_memory_stats()
        return res
    return wrapper


async def async_measure(docs: tuple[TalismanDocument, ...], model: AbstractDocumentProcessor, count: int):
    async with model:
        measure(docs, model, count)


def measure(docs: tuple[TalismanDocument, ...], model: AbstractDocumentProcessor, count: int):
    def runner():
        config_type = model.config_type
        model.process_docs(docs, config_type())

    _measure = measure_default(runner)

    try:
        _measure = measure_linux(_measure)
    except ImportError:
        logger.warning('Ignoring Linux metrics.')

    try:
        _measure = measure_torch(_measure)
    except ImportError:
        logger.warning('Ignoring CUDA metrics.')

    average_metrics = defaultdict(float)

    for run_idx in range(count):
        for metric_name, metric_value in _measure().items():
            average_metrics[metric_name] += metric_value
            logger.info(f'[Run {run_idx}] {metric_name}: {metric_value}')

    for metric_name, metric_value in average_metrics.items():
        logger.info(f'[Average] {metric_name}: {metric_value / count}')



import json
import logging
import operator
from itertools import groupby, starmap
from os import PathLike
from pathlib import Path
from pickle import PickleError
from typing import Callable, Dict, FrozenSet, Iterable, List, Tuple, Union

import yaml

from tp_interfaces.abstract import AbstractDocumentProcessor
from tp_interfaces.abstract.processor.processor import AbstractSerializableDocumentProcessor
from tp_interfaces.processors.updatable_composite import UpdatableSequentialDocumentProcessor
from tp_interfaces.serializable import load_object

ActionsType = Dict[FrozenSet[str], Callable[[dict], AbstractDocumentProcessor]]

_logger = logging.getLogger(__name__)


def _process_config(config: dict) -> AbstractDocumentProcessor:
    from talisman_tools.plugin import WrapperActionsPlugins

    plugin = config.get("plugin")
    actions = WrapperActionsPlugins.plugins[plugin]

    proper_actions = [action for required_key_set, action in actions.items() if required_key_set <= set(config.keys())]
    if len(proper_actions) > 1:
        raise Exception("Provided config can be processed by multiple actions")
    if len(proper_actions) == 0:
        raise Exception("Provided config can not be processed by any action")

    action, = proper_actions
    return action(config)


def wrap_model_from_config(config: Union[List[dict], dict], *, merge: bool = True) -> AbstractDocumentProcessor:
    if isinstance(config, list) and len(config) == 1:
        config = config[0]
    if isinstance(config, dict):
        return _process_config(config)
    if isinstance(config, list):
        if len(config) == 0:
            raise Exception("Provided list config has no elements")

        def _process_sequential_config(cfg: dict) -> Tuple[str, AbstractDocumentProcessor]:
            return cfg.get("name", ''), _process_config(cfg)

        processors = tuple(map(_process_sequential_config, config))
        if len(processors) != len(set(map(operator.itemgetter(0), processors))):
            raise ValueError(f"processor names collision")

        return UpdatableSequentialDocumentProcessor.build(processors, merge=merge)

    else:
        raise Exception("Provided config contains neither list, nor dict")


def configure_model(config: Union[List[dict], dict], *, merge: bool = True) -> AbstractDocumentProcessor:
    from talisman_tools.plugin import ProcessorPlugins  # inline import to avoid circular imports

    if isinstance(config, list) and len(config) == 1:
        config = config[0]
    if isinstance(config, dict):  # build model from config
        if config.get("plugin") is None and config["model"] == "wrapper":
            return wrap_model_from_config(config["config"], merge=merge)
        return ProcessorPlugins.plugins[config["plugin"]][config["model"]].from_config(config.get("config", {}))
    if isinstance(config, list):
        if len(config) == 0:
            raise Exception("Couldn't configure empty list model")
        key = lambda c: (c.get("plugin", ''), c["model"])
        groupped = {k: tuple(v) for k, v in groupby(config, key=key)}

        models = []

        for (plugin, model), cfgs in groupped.items():
            if plugin == '' and model == "wrapper":
                for cfg in cfgs:
                    models.append((cfg.get("name", ''), configure_model(cfg, merge=merge)))
                continue
            type_ = ProcessorPlugins.plugins[plugin][model]
            for cfg in cfgs:
                models.append((cfg.get("name", ''), type_.from_config(cfg.get("config", {}))))

        return UpdatableSequentialDocumentProcessor.build(models, merge=merge)


def load_or_configure(model_or_config_named_paths: Iterable[Tuple[str, Union[str, PathLike]]], *, merge: bool = True) \
        -> AbstractDocumentProcessor:
    def load_as_model(model_path: Path) -> AbstractDocumentProcessor:
        return load_object(model_path, expected_class=AbstractSerializableDocumentProcessor)

    def load_as_config(config_path: Path) -> AbstractDocumentProcessor:
        config = read_config(config_path)
        return configure_model(config, merge=merge)

    def convert_to_model(model_or_config_path: Path) -> AbstractDocumentProcessor:
        if model_or_config_path.is_dir() or model_or_config_path.suffix == '.pkl':
            return load_as_model(model_or_config_path)

        if model_or_config_path.suffix == '.json' or model_or_config_path.suffix == '.yaml':
            return load_as_config(model_or_config_path)

        _logger.warning(f'Unrecognized filename extension for {model_or_config_path}')

        try:
            return load_as_config(model_or_config_path)
        except ValueError:
            pass

        try:
            return load_as_model(model_or_config_path)
        except PickleError:
            pass

        raise ValueError(f'Unable to read {model_or_config_path} neither as model configuration file nor as serialized model file!')

    def convert(name: str, model_or_config_path: Union[str, PathLike]) -> Tuple[str, AbstractDocumentProcessor]:
        return name, convert_to_model(Path(model_or_config_path))

    models = tuple(starmap(convert, model_or_config_named_paths))

    if not models:
        raise ValueError("Neither model paths nor model configs were provided")

    # here we could get `Composite(Composite(...), Composite(...))`
    return UpdatableSequentialDocumentProcessor.build(models, merge=merge)


def read_config(path: Union[str, Path]):
    with open(path, "r", encoding="utf-8") as f:
        if Path(path).suffix[1:] == "yaml":
            return yaml.safe_load(f)
        return json.load(f)



from argparse import ArgumentParser, Namespace
from typing import Callable

from talisman_tools.commands.servers.arguments.server import uvicorn_server_factory
from talisman_tools.parser_helper import configure_subparsers


def configure_server_parser(parser: ArgumentParser) -> None:
    from .commands import SUBPARSERS

    launcher = uvicorn_server_factory(parser)

    def get_action() -> Callable[[Namespace], None]:
        def action(args: Namespace) -> None:
            if not hasattr(args, 'server_action'):
                parser.print_help()
                exit(1)

            fast_api = args.server_action()(args)
            launcher(fast_api, args)

        return action

    parser.set_defaults(action=get_action)
    configure_subparsers(parser, SUBPARSERS)



from typing import Callable, Type

from tp_interfaces.domain.interfaces import DomainProducer
from .abstract import AbstractPluginManager


def _get_stub() -> Type[DomainProducer]:
    from tp_interfaces.domain.stub import StubDomainProducer
    return StubDomainProducer


class DomainPluginManager(AbstractPluginManager):
    def __init__(self):
        AbstractPluginManager.__init__(self, 'DOMAIN_FACTORY', _get_stub)

    @staticmethod
    def _check_value(value) -> bool:
        return isinstance(value, Callable)



import logging
import logging.config
from contextlib import AbstractContextManager
from typing import Any, AsyncContextManager, Dict, Generator, Sequence, Tuple

from fastapi import FastAPI, Request
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from pydantic import PydanticTypeError
from pydantic.error_wrappers import ErrorWrapper, ValidationError
from starlette import status
from starlette.responses import JSONResponse

_logger = logging.getLogger(__name__)


def register_context_manager(app: FastAPI, context_manager: AbstractContextManager):
    @app.on_event("startup")
    async def enter_manager():
        context_manager.__enter__()

    @app.on_event("shutdown")
    async def exit_manager():
        context_manager.__exit__(None, None, None)


def async_register_context_manager(app: FastAPI, context_manager: AsyncContextManager):
    @app.on_event("startup")
    async def enter_manager():
        await context_manager.__aenter__()

    @app.on_event("shutdown")
    async def exit_manager():
        await context_manager.__aexit__(None, None, None)


def register_exception_handlers(app: FastAPI, logger: logging.Logger = _logger):
    @app.exception_handler(RequestValidationError)
    async def validation_exception_handler(request, exc: RequestValidationError):
        error_content = jsonable_encoder({"detail": list(unwrap(exc.raw_errors))})
        logger.error("Pydantic validation error", extra=error_content)
        return JSONResponse(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            content=error_content,
        )

    def unwrap(errors: Sequence[Any]) -> Generator[Dict[str, Any], None, None]:
        for error in errors:
            if isinstance(error, ErrorWrapper):
                yield wrap_description(error)
            elif isinstance(error, list):
                yield from unwrap(error)
            else:
                logging.error(f'Unknown error object: {error}')
                raise RuntimeError(f'Unknown error object: {error}')

    def wrap_description(wrap: ErrorWrapper) -> Dict[str, Any]:
        loc = wrap.loc_tuple()  # validated parameter
        if isinstance(wrap.exc, ValidationError):
            model, errors = validation_error(wrap.exc)
        elif isinstance(wrap.exc, PydanticTypeError):
            model, errors = type_error(wrap.exc)
        else:
            model, errors = type(wrap.exc).__name__, []
        return {'model': model, 'loc': loc, 'errors': errors}

    def validation_error(exc: ValidationError) -> Tuple[str, list]:
        return exc.model.__name__, exc.errors()

    def type_error(exc: PydanticTypeError) -> Tuple[str, str]:
        return type(exc).__name__, exc.msg_template


def log_debug_data(logger: logging.Logger, msg: str, request: Request = None, **kwargs) -> None:
    extras = {}
    if request is not None:
        extras['client'] = request.client
    extras.update({key: jsonable_encoder(value) for key, value in kwargs.items()})
    logger.debug(msg, extra=extras)



from argparse import ArgumentParser
from typing import Callable, Dict, Tuple

from .converter import configure_converter_parser
from .loader import configure_loader_parser
from .processor import configure_processor_parser

# Every subcommand should have its own parser with description
# Note that subparsers should change the default `server_action` argument
SUBPARSERS: Dict[str, Tuple[str, Callable[[ArgumentParser], None]]] = {
    'loader': ('Configure loader server', configure_loader_parser),
    'processor': ('Configure processor server', configure_processor_parser),
    'converter': ('Configure document converter', configure_converter_parser)
}



import math
from collections import defaultdict
from itertools import chain, starmap
from statistics import harmonic_mean, mean
from typing import Callable, Dict, Hashable, Iterable, List, NamedTuple, Sequence, Set, Tuple, Union

import numpy as np
from scipy.optimize import linear_sum_assignment


# copy of derek.common.evaluation.metrics


def f1_score(precision, recall):
    denominator = precision + recall
    return 2 * precision * recall / denominator if denominator > 0 else 0


def binary_precision_score(predicted: List[bool], gold: List[bool]):
    fp, fn, tp, tn = _compute_binary_errors(predicted, gold)
    denominator = tp + fp
    return tp / denominator if denominator > 0 else 0


def binary_recall_score(predicted: List[bool], gold: List[bool]):
    fp, fn, tp, tn = _compute_binary_errors(predicted, gold)
    denominator = tp + fn
    return tp / denominator if denominator > 0 else 0


def binary_f1_score(predicted: List[bool], gold: List[bool]):
    precision = binary_precision_score(predicted, gold)
    recall = binary_recall_score(predicted, gold)
    return f1_score(precision, recall)


def binary_accuracy(predicted: List[bool], gold: List[bool]):
    fp, fn, tp, tn = _compute_binary_errors(predicted, gold)
    denominator = tp + tn + fp + fn
    return (tp + tn) / denominator if denominator > 0 else 0


def binary_macro_avg_score(score_func, segments_predicted: List[List[bool]], segments_gold: List[List[bool]]):
    score = sum(score_func(pred, gold) for pred, gold in zip(segments_predicted, segments_gold))
    return score / len(segments_predicted) if len(segments_predicted) > 0 else 0


def binary_micro_avg_score(score_func, segments_predicted: List[List[bool]], segments_gold: List[List[bool]]):
    return score_func(list(chain.from_iterable(segments_predicted)), list(chain.from_iterable(segments_gold)))


def binary_macro_scores(segments_predicted: List[List[bool]], segments_gold: List[List[bool]]):
    macro_precision = binary_macro_avg_score(binary_precision_score, segments_predicted, segments_gold)
    macro_recall = binary_macro_avg_score(binary_recall_score, segments_predicted, segments_gold)
    macro_f1 = binary_macro_avg_score(binary_f1_score, segments_predicted, segments_gold)

    return macro_precision, macro_recall, macro_f1


def binary_micro_scores(segments_predicted: List[List[bool]], segments_gold: List[List[bool]]):
    micro_precision = binary_micro_avg_score(binary_precision_score, segments_predicted, segments_gold)
    micro_recall = binary_micro_avg_score(binary_recall_score, segments_predicted, segments_gold)
    micro_f1 = binary_micro_avg_score(binary_f1_score, segments_predicted, segments_gold)

    return micro_precision, micro_recall, micro_f1


def _compute_binary_errors(predicted: List[bool], gold: List[bool]):
    fp, fn, tp, tn = 0, 0, 0, 0

    for sample_pred, sample_gold in zip(predicted, gold):
        if sample_pred and sample_gold:
            tp += 1
        if sample_pred and not sample_gold:
            fp += 1
        if not sample_pred and sample_gold:
            fn += 1
        if not sample_pred and not sample_gold:
            tn += 1

    return fp, fn, tp, tn


def ir_precision_score(predicted: set, gold: set):
    return len(predicted.intersection(gold)) / len(predicted) if len(predicted) > 0 else 0


def ir_recall_score(predicted: set, gold: set):
    return len(predicted.intersection(gold)) / len(gold) if len(gold) > 0 else 0


def ir_f1_score(predicted: set, gold: set):
    intersection = predicted.intersection(gold)
    precision = len(intersection) / len(predicted) if len(predicted) > 0 else 0
    recall = len(intersection) / len(gold) if len(gold) > 0 else 0

    return f1_score(precision, recall)


def ir_macro_avg_score(score_func, segments_predicted: List[set], segments_gold: List[set]):
    score = 0

    for segment_predicted, segment_gold in zip(segments_predicted, segments_gold):
        score += score_func(segment_predicted, segment_gold)

    return score / len(segments_predicted) if len(segments_predicted) > 0 else 0


def ir_macro_scores(segments_predicted: List[set], segments_gold: List[set]):
    macro_precision = ir_macro_avg_score(ir_precision_score, segments_predicted, segments_gold)
    macro_recall = ir_macro_avg_score(ir_recall_score, segments_predicted, segments_gold)
    macro_f1 = ir_macro_avg_score(ir_f1_score, segments_predicted, segments_gold)

    return macro_precision, macro_recall, macro_f1


def ir_micro_avg_score(score_func, segments_predicted: List[set], segments_gold: List[set]):
    predicted = {(i, pred) for i, segment_pred in enumerate(segments_predicted) for pred in segment_pred}
    gold = {(i, gold) for i, segment_gold in enumerate(segments_gold) for gold in segment_gold}

    return score_func(predicted, gold)


def ir_micro_scores(segments_predicted: List[set], segments_gold: List[set]):
    predicted = {(i, pred) for i, segment_pred in enumerate(segments_predicted) for pred in segment_pred}
    gold = {(i, gold) for i, segment_gold in enumerate(segments_gold) for gold in segment_gold}

    micro_precision = ir_precision_score(predicted, gold)
    micro_recall = ir_recall_score(predicted, gold)
    micro_f1 = ir_f1_score(predicted, gold)

    return micro_precision, micro_recall, micro_f1


def ir_categorized_queries_score(
        queries: List,
        queries_predicted: List[set], queries_gold: List[set],
        categorizer: Callable, score_func: Callable):

    queries_predicted_filtered = []
    queries_gold_filtered = []
    possible_categories = set()

    for query, query_pred, query_gold in zip(queries, queries_predicted, queries_gold):
        filtered_pred = _categorize_query_objects(query, query_pred, categorizer)
        filtered_gold = _categorize_query_objects(query, query_gold, categorizer)

        queries_predicted_filtered.append(filtered_pred)
        queries_gold_filtered.append(filtered_gold)
        possible_categories.update(filtered_gold.keys(), filtered_pred.keys())

    categories_results = {}

    for category in possible_categories:
        # defaultdict provide empty set if category is not present in query
        category_predicted = [query_pred[category] for query_pred in queries_predicted_filtered]
        category_gold = [query_gold[category] for query_gold in queries_gold_filtered]

        category_score = score_func(category_predicted, category_gold)
        categories_results[category] = category_score

    return categories_results


def _categorize_query_objects(query, objects: set, categorizer: Callable) -> defaultdict:
    filtered_objects = defaultdict(set)

    for obj in objects:
        filtered_objects[categorizer(query, obj)].add(obj)

    return filtered_objects


def reciprocal_rank(relevant: Union[set, frozenset], ranked_responses: Sequence) -> float:
    for i, obj in enumerate(ranked_responses, start=1):
        if obj in relevant:
            return 1 / i
    return 0.0


def mean_reciprocal_rank(segments_predicted: Sequence[Sequence], segments_gold: Sequence[Union[set, frozenset]]) -> float:
    if len(segments_predicted) != len(segments_gold):
        raise ValueError("Segments must be aligned")

    macro_sum = sum(starmap(reciprocal_rank, zip(segments_gold, segments_predicted)), 0.0)
    return macro_sum / len(segments_predicted) if len(segments_predicted) else 0.0


# following this are coreference metrics adapted from scorch 0.2.0

def _trace(cluster: Set, partition: Iterable[Set]) -> Iterable[Set]:
    r"""
    Return the partition of `#cluster` induced by `#partition`, that is
    ```math
    \{C∩A|A∈P\} ∪ \{\{x\}|x∈C∖∪P\}
    ```
    Where `$C$` is `#cluster` and `$P$` is `#partition`.
    This assume that the elements of `#partition` are indeed pairwise disjoint.
    """
    remaining = set(cluster)
    for a in partition:
        common = remaining.intersection(a)
        if common:
            remaining.difference_update(common)
            yield common
    for x in sorted(remaining):
        yield {x}


class RemapClusteringsReturn(NamedTuple):
    clusterings: Sequence[Sequence[Sequence[int]]]
    elts_map: Dict[Hashable, int]


def remap_clusterings(
    clusterings: Sequence[Sequence[Set[Hashable]]],
) -> RemapClusteringsReturn:
    """Remap clusterings of arbitrary elements to clusterings of integers."""
    elts = {e for clusters in clusterings for c in clusters for e in c}
    elts_map = {e: i for i, e in enumerate(elts)}
    res = []
    for clusters in clusterings:
        remapped_clusters = []
        for c in clusters:
            remapped_c = [elts_map[e] for e in c]
            remapped_clusters.append(remapped_c)
        res.append(remapped_clusters)
    return RemapClusteringsReturn(res, elts_map)


def muc_score(key: Sequence[Set], response: Sequence[Set]) -> Tuple[float, float, float]:
    r"""
        Compute the MUC `$(R, P, F₁)$` scores for a `#response` clustering given a `#key` clustering,
        that is
        ```math
        R &= \frac{∑_{k∈K}(\#k-\#p(k, R))}{∑_{k∈K}(\#k-1)}\\
        P &= \frac{∑_{r∈R}(\#r-\#p(r, K))}{∑_{r∈R}(\#r-1)}\\
        F &= 2*\frac{PR}{P+R}
        ```
        with `$p(x, E)=\{x∩A|A∈E\}$`.
        In the edge case where all clusters in either `#key` or `#response` are singletons, `$P$`, `$R$`
        and `$F$` are defined to be `$0$`, following the reference implementation (since singleton
        clusters where not considered in Vilain et al. (1995).
        Note: This implementation is significantly different from the reference one (despite
        implementing the formulae from Pradahan et al. (2014) in that the reference use the ordering of
        mentions in documents to consistently assign a non-problematic spanning tree (viz. a chain) to
        each cluster, thus avoiding the issues that led Vilain et al. (1995) to define MUC by the
        formulae above.
        """
    # Edge case
    if all(len(k) == 1 for k in key) or all(len(r) == 1 for r in response):
        return 0.0, 0.0, 0.0

    def func(one: Sequence[Set], another: Sequence[Set]):
        return sum(len(o) - sum(1 for _ in _trace(o, another)) for o in one) / sum(len(o) - 1 for o in one)

    r = func(key, response)
    p = func(response, key)
    f = harmonic_mean((float(r), float(p)))
    return r, p, f


def bcub_score(
    key: Sequence[Set], response: Sequence[Set]
) -> Tuple[float, float, float]:
    r"""
    Compute the B³ `$(R, P, F₁)$` scores for a `#response` clustering given a `#key` clustering,
    that is
    ```math
    R &= \frac{∑_{k∈K}∑_{r∈R}\frac{(\#k∩r)²}{\#k}}{∑_{k∈K}\#k}\\
    P &= \frac{∑_{r∈R}∑_{k∈K}\frac{(\#r∩k)²}{\#r}}{∑_{r∈R}\#r}\\
    F &= 2*\frac{PR}{P+R}
    ```
    """

    def func(one, another):
        if sum(map(len, one)) == 0:
            return 0.0
        return math.fsum(len(o.intersection(a)) ** 2 / len(o) for o in one for a in another) / sum(len(o) for o in one)

    r = func(key, response)
    p = func(response, key)
    f = harmonic_mean((float(r), float(p)))
    return r, p, f


def ceaf_score(
    key: Sequence[Set],
    response: Sequence[Set],
    score: Callable[[Set, Set], float],
) -> Tuple[float, float, float]:
    r"""
    Compute the CEAF `$(R, P, F₁)$` scores for a `#response` clustering given a `#key` clustering
    using the `#score` alignment score function, that is
    ```math
    R &= \frac{∑_{k∈K}C(k, A(k))}{∑_{k∈K}C(k, k)}\\
    P &= \frac{∑_{r∈R}C(r, A⁻¹(r))}{∑_{r∈R}C(r, r)}\\
    F &= 2*\frac{PR}{P+R}
    ```
    Where `$C$` is `#score` and `$A$` is a one-to-one mapping from key clusters to response
    clusters that maximizes `$∑_{k∈K}C(k, A(k))$`.
    """
    if len(response) == 0 or len(key) == 0:
        return 0.0, 0.0, 0.0
    else:
        cost_matrix = np.array([[-score(k, r) for r in response] for k in key])
        # TODO: See https://github.com/allenai/allennlp/issues/2946 for ideas on speeding
        # the next line up
        row_ind, col_ind = linear_sum_assignment(cost_matrix)
        total_score = -cost_matrix[row_ind, col_ind].sum()
        r = total_score / math.fsum(score(k, k) for k in key)
        p = total_score / math.fsum(score(res, res) for res in response)
        f = harmonic_mean((float(r), float(p)))
        return r, p, f


def ceafe_score(
    key: Sequence[Set], response: Sequence[Set]
) -> Tuple[float, float, float]:
    r"""
    Compute the CEAFₑ `$(R, P, F₁)$` scores for a `#response` clustering given a `#key`
    clustering, that is the CEAF score for the `$Φ₄$` score function (aka the Sørensen–Dice
    coefficient).
    ```math
    Φ₄: (k, r) ⟼ \frac{2×\#k∩r}{\#k+\#r}
    ```
    Note: this use the original (Luo, 2005) definition as opposed to Pradhan et al. (2014)'s one
    which inlines the denominators.
    """

    def theta_4(k, r):
        return 2 * len(k.intersection(r)) / (len(k) + len(r))

    return ceaf_score(key, response, theta_4)


def conll2012_score(key: Sequence[Set], response: Sequence[Set]) -> float:
    r"""
    Return the CoNLL-2012 scores for a `#response` clustering given a `#key` clustering, that is,
    the average of the MUC, B³ and CEAFₑ scores.
    """
    return mean((metric(key, response)[2] for metric in (muc_score, bcub_score, ceafe_score)))

# lea section adapted from CoVal at github.com/ns-moosavi/coval


def get_mention_assignments(in_clusters, out_clusters):
    mention_cluster_ids = {}
    res = {}
    for ix, cluster in enumerate(out_clusters):
        for mention in cluster:
            res[mention] = ix

    for cluster in in_clusters:
        for mention in cluster:
            if mention in res:
                mention_cluster_ids[mention] = res[mention]

    return mention_cluster_ids


def lea_score(key: Sequence[Set], response: Sequence[Set]):
    listify = lambda seq: [list(x) for x in seq]
    lkey = listify(key)
    lresponse = listify(response)
    key_mention_assignment = get_mention_assignments(lkey, lresponse)
    res_mention_assignment = get_mention_assignments(lresponse, lkey)
    pn, pd = lea(lresponse, lkey, res_mention_assignment)
    rn, rd = lea(lkey, lresponse, key_mention_assignment)
    p = 0 if pd == 0 else pn / float(pd)
    r = 0 if rd == 0 else rn / float(rd)
    f1 = 0 if p + r == 0 else 2 * p * r / (p + r)

    return r, p, f1


def lea(input_clusters, output_clusters, mention_to_gold):
    num, den = 0, 0

    for c in input_clusters:
        if len(c) == 1:
            all_links = 1
            if c[0] in mention_to_gold and len(
                    output_clusters[mention_to_gold[c[0]]]) == 1:
                common_links = 1
            else:
                common_links = 0
        else:
            common_links = 0
            all_links = len(c) * (len(c) - 1) / 2.0
            for i, m in enumerate(c):
                if m in mention_to_gold:
                    for m2 in c[i + 1:]:
                        if m2 in mention_to_gold and mention_to_gold[
                                m] == mention_to_gold[m2]:
                            common_links += 1

        num += len(c) * common_links / float(all_links)
        den += len(c)

    return num, den



import os


def get_max_concurrent_threads(*, default: int = 20) -> int:
    max_threads = os.getenv("MAX_CONCURRENT_THREADS")
    return int(max_threads) if max_threads else default


def get_max_treated_jobs() -> int:
    max_treated_jobs = os.environ.get('MAX_TREATED_JOBS')
    return int(max_treated_jobs) if max_treated_jobs is not None else 5



from tp_interfaces.abstract import ModelTypeFactory
from .abstract import AbstractPluginManager


class ProcessorPluginManager(AbstractPluginManager):
    def __init__(self):
        AbstractPluginManager.__init__(self, 'PROCESSORS')

    @staticmethod
    def _check_value(value) -> bool:
        return isinstance(value, ModelTypeFactory)



from talisman_tools.configure import configure_model
from tp_interfaces.abstract import AbstractDocumentProcessor


def wrap_model_action(config: dict) -> AbstractDocumentProcessor:
    """
    create model from config and wraps it with respect to specified wrapper and wrapper configuration
    :param config: dict with required "wrapper" and "to_wrap" keys. Optional "plugin" and "config" keys are taken into account
    :return: wrapped document processor
    """
    from talisman_tools.plugin import WrapperPlugins  # inline import to avoid circular imports
    if isinstance(config["wrapper"], dict):
        plugin = config["wrapper"].get("plugin")
        model = config["wrapper"]["model"]
    elif isinstance(config["wrapper"], str):
        plugin = None
        model = config["wrapper"]
    else:
        raise ValueError
    try:
        wrapper_factory = WrapperPlugins.plugins[plugin][model]
        return wrapper_factory.from_config(configure_model(config['to_wrap']), config.get("config", {}))
    except ValueError:
        raise ValueError(f"{config['wrapper']} is not found in {config.get('plugin') or 'default'} plugin")


def configure_model_action(config: dict) -> AbstractDocumentProcessor:
    """
    create model from config
    :param config: dict with required "model" key
    :return: configured model
    """
    return configure_model(config)



from typing import Callable, Dict, List, Sequence, Tuple

from tdm import TalismanDocument
from tdm.datamodel.facts import ConceptFact

from talisman_tools.commands.model.commands.evaluate.metrics import ir_micro_avg_score, ir_recall_score, mean_reciprocal_rank


def _compute_micro_recall_at_k(predicted: List[tuple], gold: List[set], k: int) -> float:
    if k <= 0:
        raise ValueError(f"{k} must be > 0")

    predicted_at_k = [set(s[:k]) for s in predicted]
    return ir_micro_avg_score(ir_recall_score, predicted_at_k, gold)


_Predicate = Callable[[ConceptFact], bool]


def _truth_predicate(fact: ConceptFact) -> bool:
    return True


def _in_kb_predicate(fact: ConceptFact) -> bool:
    return fact.value is not None and fact.value


def _ambiguity_predicate(fact: ConceptFact) -> bool:
    return fact.value is not None and isinstance(fact.value, tuple) and len(fact.value) >= 2


def _filter_with_mode(predicted: Sequence[TalismanDocument], gold: Sequence[TalismanDocument], only_in_kb: bool, mode: str) -> \
        Tuple[Sequence[TalismanDocument], Sequence[TalismanDocument]]:

    if mode == "default" and not only_in_kb:
        return predicted, gold

    gold_predicate = _in_kb_predicate if only_in_kb else _truth_predicate
    if mode == "default":
        pred_predicate = _truth_predicate
    elif mode == "ambigous":
        pred_predicate = _ambiguity_predicate
    elif mode == "unambigous":
        pred_predicate = lambda fact: not _ambiguity_predicate(fact)
    else:
        raise ValueError(f"{mode} is not supported")

    pred_filtered, gold_filtered = [], []
    for pred_doc, gold_doc in zip(predicted, gold):
        pred_facts = tuple(pred_doc.filter_facts(ConceptFact, pred_predicate))
        gold_facts = tuple(gold_doc.filter_facts(ConceptFact, gold_predicate))

        mentions_intersection = {fact.mention for fact in pred_facts}.intersection(fact.mention for fact in gold_facts)

        pred_filtered.append(pred_doc.without_facts().with_facts(fact for fact in pred_facts if fact.mention in mentions_intersection))
        gold_filtered.append(gold_doc.without_facts().with_facts(fact for fact in gold_facts if fact.mention in mentions_intersection))

    return pred_filtered, gold_filtered


def is_ambiguous(doc: TalismanDocument) -> bool:  # TODO: move it to helpers/tdm
    return any(isinstance(fact.value, tuple) and fact.value for fact in doc.facts[ConceptFact])


def _group_facts(doc: TalismanDocument) -> Dict[tuple, ConceptFact]:
    result = {}
    for fact in doc.facts[ConceptFact]:
        key = fact.type_id, fact.mention
        if key in result:
            raise ValueError
        result[key] = fact
    return result


def evaluate_dmb(
        predicted: Sequence[TalismanDocument], gold: Sequence[TalismanDocument], eval_config: dict, at_ks: Sequence[int],
        only_in_kb: bool = False, mode: str = "default") -> Dict[str, float]:
    """
    :param predicted: Sequence of disambiguated Documents with ranked concept fact values ranked_concepts
    :param gold: Sequence of unambiguous Documents with gold concept fact values
    :param eval_config: optional evaluation config
    :param at_ks: Compute micro_recall@k for all k from at_ks
    :param only_in_kb: Evaluate only in KB mentions
    :param mode: "default" | "ambiguous" | "non_ambigous".
        If "default", evaluate all mentions.
        If "ambigous" evaluate only ambigous mentions (with at least 2 predicted concepts)
        If "unambigous" evaluate only unambigous mentions (with at most 1 predicted concept)
    :return: dictionary with keys:
        "num_mentions" -- number of evaluated mentions
        "micro_recall@k" -- value of micro_recall at k for each k from at_ks param
        "mean_reciprocal_rank" -- value of mean reciprocal rank
    """
    if len(predicted) != len(gold):
        raise ValueError("Predicted and gold sequences are not aligned")
    if any(map(is_ambiguous, gold)):
        raise ValueError("Gold sequence contains ambiguous documents")

    predicted, gold = _filter_with_mode(predicted, gold, only_in_kb, mode)

    def response_for_gold(fact: ConceptFact) -> set:
        return {fact.value}

    def response_for_pred(fact: ConceptFact) -> tuple:
        if fact is None:
            return (None, )
        if isinstance(fact.value, tuple):
            return fact.value
        return (fact.value, )  # None or single concept id

    gold_responses = []
    predicted_responses = []
    for gold_doc, pred_doc in zip(gold, predicted):
        if gold_doc.without_facts() != pred_doc.without_facts():
            raise ValueError("Predicted and gold sequences are not aligned")

        gold_facts = _group_facts(gold_doc)
        pred_facts = _group_facts(pred_doc)

        for key, gold_fact in gold_facts.items():  # mentions are aligned
            gold_responses.append(response_for_gold(gold_fact))
            predicted_responses.append(response_for_pred(pred_facts.get(key)))

    ret = {f"micro_recall@{k}": _compute_micro_recall_at_k(predicted_responses, gold_responses, k) for k in at_ks}
    ret["mean_reciprocal_rank"] = mean_reciprocal_rank(predicted_responses, gold_responses)
    ret["num_mentions"] = len(gold_responses)
    return ret



from argparse import ArgumentParser, Namespace

from talisman_tools.plugin import DomainPlugins
from tp_interfaces.domain.manager import DomainManager
from tp_interfaces.helpers.io import read_json


def get_domain_factory(parser: ArgumentParser):
    parser.add_argument('-domain_config_path', metavar='<domain config path>')

    def set_domain(args: Namespace) -> None:
        if args.domain_config_path is None:
            return

        config = read_json(args.domain_config_path)
        domain_producer = DomainPlugins.plugins[config.get("plugin")]().from_config(config.get('config', {}))
        DomainManager().set_producer(domain_producer)

    return set_domain



import asyncio
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Callable

from talisman_tools.arguments.domain import get_domain_factory
from talisman_tools.arguments.reader import get_reader_factory
from talisman_tools.arguments.serializer import get_serializer_factory
from tp_interfaces.abstract import AbstractDocumentProcessor


def configure_process_parser(parser: ArgumentParser) -> None:
    domain_factory = get_domain_factory(parser)
    reader_factory = get_reader_factory(parser, {'docs_path': '<docs path>'})
    serializer_factory = get_serializer_factory(parser)

    parser.add_argument('--config', type=Path, metavar='<document processing config path>')
    parser.add_argument('--batch', type=int, metavar='<batch size>', default=1000)

    def get_action() -> Callable[[AbstractDocumentProcessor, Namespace], None]:
        def action(processor: AbstractDocumentProcessor, args: Namespace) -> None:
            domain_factory(args)
            reader = reader_factory(args)['docs_path']
            serializer = serializer_factory(args)
            from .action import async_action
            return asyncio.run(async_action(processor, serializer, reader, args))

        return action

    parser.set_defaults(processor_action=get_action)



from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Tuple

from tp_interfaces.abstract import AbstractDocumentProcessor


def _named_path(arg: str) -> Tuple[str, Path]:
    if ":" not in arg:
        return "", Path(arg)
    name, path = arg.split(":", 1)
    return name, Path(path)


def get_processor_factory(parser: ArgumentParser):
    argument_group = parser.add_argument_group(title='Model arguments')
    argument_group.add_argument('-m', '--model_paths', action='append', type=_named_path, metavar='<model name:model path>', required=True,
                                help='Named paths to configuration files or serialized models')
    argument_group.add_argument('--wrapper', type=Path, metavar='<wrapper path>', help='Path to wrapper configuration file', default=None)
    argument_group.add_argument('--omit_merging', action='store_true', help='Disable model merging during configuration')

    def get_processor(args: Namespace) -> AbstractDocumentProcessor:
        from talisman_tools.configure import load_or_configure, wrap_model
        from talisman_tools.configure.configure import read_config

        processor = load_or_configure(model_or_config_named_paths=args.model_paths, merge=not args.omit_merging)
        if args.wrapper is not None:
            processor = wrap_model(processor, read_config(args.wrapper), merge=not args.omit_merging)
        return processor

    return get_processor



from dataclasses import dataclass
from functools import partial
from typing import Iterable, Iterator

from tdm import TalismanDocument
from tdm.datamodel.facts import MentionFact
from tdm.datamodel.mentions import TextNodeMention
from typing_extensions import Self

from talisman_tools.commands.model.commands.evaluate.evaluation import _evaluate
from tp_interfaces.helpers.datamodel.domain import get_id


@dataclass(frozen=True, slots=True)
class MentionTupleView:
    """"
    A utility class that represents mention facts as named tuples (without fact ids)
    """

    type_id: str

    node_id: str
    start: int
    end: int

    @classmethod
    def from_fact(cls, fact: MentionFact) -> Self:
        if not isinstance(fact.mention, TextNodeMention):
            raise ValueError(f'Cannot create {cls.__class__.__name__} from fact with mention of class {fact.mention.__class__.__name__}!')
        return cls(get_id(fact.value.type_id), fact.mention.node_id, fact.mention.start, fact.mention.end)

    @classmethod
    def from_doc(cls, doc: TalismanDocument, excluded_types: dict) -> Iterator[Self]:
        values_blacklist = set(excluded_types.get('values', []))

        for fact in doc.get_facts(MentionFact, filter_=lambda m: isinstance(m.mention, TextNodeMention)):
            if get_id(fact.value.type_id) not in values_blacklist:
                yield cls.from_fact(fact)


def _fact_type(query, fact: MentionTupleView) -> str:
    return fact.type_id


_NERC_CATEGORIZERS = {
    'fact_type': _fact_type
}


def evaluate_nerc(predicted: Iterable[TalismanDocument], gold: Iterable[TalismanDocument], eval_config: dict) -> dict[str, dict]:
    return _evaluate(
        predicted, gold, partial(MentionTupleView.from_doc, excluded_types=eval_config.get('excluded_types', {})), _NERC_CATEGORIZERS
    )



from argparse import ArgumentParser, Namespace
from typing import Callable

from fastapi import FastAPI

from talisman_tools.arguments.domain import get_domain_factory


def configure_converter_parser(parser: ArgumentParser) -> None:
    domain_factory = get_domain_factory(parser)

    def get_action() -> Callable[[Namespace], FastAPI]:
        def action_with_extra(args: Namespace) -> FastAPI:
            from .action import action
            domain_factory(args)
            return action(args)

        return action_with_extra

    parser.set_defaults(server_action=get_action)



__all__ = [
    'DEFAULT_WRAPPER_ACTIONS'
]

from .actions import configure_model_action, wrap_model_action

DEFAULT_WRAPPER_ACTIONS = {
    frozenset(("wrapper", "to_wrap")): wrap_model_action,
    frozenset(("model", )): configure_model_action
}



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("VERSION", "r", encoding="utf-8") as f:
    version = f.read()

setup(
    name='talisman-tools',
    version=version,
    description='Talisman applications',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='ISPRAS Talisman NLP team',
    author_email='modis@ispras.ru',
    maintainer='Vladimir Mayorov',
    maintainer_email='vmayorov@ispras.ru',
    packages=find_packages(include=['talisman_tools', 'talisman_tools.*']),
    install_requires=[
        'talisman-interfaces~=0.5.3', 'talisman-dm~=1.0.0a37',
        'fastapi>=0.73.0', 'pydantic~=1.10.4', 'uvicorn>=0.13.3', 'requests~=2.31.0', 'urllib3~=2.2.0',
        'jsonformatter>=0.3.0',
        'starlette', 'anyio',
        'pyyaml'
    ],
    extras_require={
        'logstash': ['python3-logstash>=0.4.80'],
        'evaluate': ['numpy', 'scipy', 'typing_extensions~=4.8.0'],
        'cuda_metrics': ['torch'],
        "tests": ['parameterized~=0.9.0']
    },
    entry_points={
        'console_scripts': [
            'talisman-tools = talisman_tools.talisman:main',
        ]
    },
    data_files=[('', ['VERSION'])],
    python_requires='>=3.10',
    license='Apache Software License',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License'
    ]
)



import logging
from logging import Logger
from typing import Optional

from fastapi import FastAPI, Request
from tdm import TalismanDocumentModel
from tdm.v0 import TalismanDocumentModel as LegacyTalismanDocumentModel

from tp_interfaces.abstract import ImmutableBaseModel
from tp_interfaces.domain.manager import DomainManager
from tp_interfaces.domain.model.types import DocumentType, PropertyType
from tp_interfaces.helpers.batch import AbstractModelInput

_logger = logging.getLogger(__name__)

_metadata_map = {
    'title': 'Заголовок',
    'file_name': 'Название файла',
    'file_type': 'Формат файла',
    'size': 'Размер файла',
    'created_time': 'Дата создания',
    'access_time': 'Дата последнего просмотра',
    'modifired_time': 'Дата изменения',
    'publication_date': 'Дата публикации',
    'publication_author': 'Автор публикации',
    'description': 'Примечание',
    'url': 'Источник',
    'access_level': 'Уровень доступа',
    'user': 'Автор документа',
    'path': 'Путь до файла',
    'trust_level': 'Уровень доверия',
    'markers': 'Метки',
    'preview_text': 'Предпросмотр текста',
    'platform': 'Платформа',
    'account': 'Аккаунт',
    'story': 'Сюжет',
    'language': 'Язык документа',
    'job_id': 'Запуск сбора',
    'periodic_job_id': 'Периодический запуск сбора',
    'task_id': 'Задача импорта',
    'periodic_task_id': 'Периодическая задача импорта',
    'parent_uuid': 'Родительский документ'
}


def register_convert_doc(
        app: FastAPI,
        logger: Logger = _logger
):
    class ModelInput(AbstractModelInput):
        message: LegacyTalismanDocumentModel
        config: ImmutableBaseModel

        def get_message(self) -> LegacyTalismanDocumentModel:
            return self.message

        def get_config(self) -> Optional[ImmutableBaseModel]:
            return self.config

    @app.post('/', response_model=TalismanDocumentModel, response_model_exclude_none=True)
    async def convert(request: Request, *, message: ModelInput):
        async with DomainManager() as manager:
            domain = await manager.domain

        titles = tuple(domain.get_types(PropertyType, filter_=lambda p: p.name == 'Название'))
        doc_type = domain.types[DocumentType].pop().id

        message.message.set_domain(domain)
        document = message.message.to_doc(doc_type, titles, _metadata_map)
        logger.debug(f"document {document.id} successfully converted")
        result = TalismanDocumentModel.serialize(document)
        return result



from argparse import ArgumentParser, Namespace
from typing import Callable

from fastapi import FastAPI

from talisman_tools.arguments.domain import get_domain_factory
from talisman_tools.arguments.loader import get_loader_factory


def configure_loader_parser(parser: ArgumentParser) -> None:
    loader_factory = get_loader_factory(parser)
    domain_factory = get_domain_factory(parser)

    def get_action() -> Callable[[Namespace], FastAPI]:
        def action_with_extra(args: Namespace) -> FastAPI:
            from .action import action
            domain_factory(args)
            loader = loader_factory(args)
            return action(args, loader)

        return action_with_extra

    parser.set_defaults(server_action=get_action)



from importlib.metadata import entry_points
from types import ModuleType
from typing import Any, Dict, Optional, TypeVar

GROUP = "talisman.plugins"


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class AbstractPluginManager(metaclass=Singleton):
    def __init__(self, attr: str, default=None):
        self._registered_plugins = {}
        if default is not None:
            self._registered_plugins[None] = default

        for entry_point in entry_points(group=GROUP):
            name = entry_point.name
            module = entry_point.load()
            if not isinstance(module, ModuleType):
                continue
            try:
                value = getattr(module, attr)
                if self._check_value(value):
                    self._registered_plugins[name] = value
            except AttributeError:
                pass

        self._disabled_plugins = set()

    @property
    def plugins(self) -> Dict[Optional[str], Any]:
        return {plugin: value for plugin, value in self._registered_plugins.items() if plugin not in self._disabled_plugins}

    @staticmethod
    def _check_value(value) -> bool:
        return True

    def disable_plugin(self, plugin: str):
        self._disabled_plugins.add(plugin)

    def enable_plugin(self, plugin: str):
        self._disabled_plugins.discard(plugin)


_T = TypeVar('_T')


def flattened(data: Dict[str, Dict[str, _T]], *, delimiter: str = ':', start: str = '') -> Dict[str, _T]:
    result = {}
    for plugin, readers in data.items():
        for key, reader in readers.items():
            result[f"{start}{plugin}{delimiter}{key}"] = reader
    return result



import logging
import os
from argparse import ArgumentParser, Namespace

from tp_interfaces.logging.context import get_log_extras

SERVICE_NAME = os.getenv('SERVICE_NAME', default='')


class TalismanFilter(logging.Filter):
    def __init__(self):
        super().__init__()
        self._service = SERVICE_NAME

    def filter(self, record) -> bool:
        if self._service:
            record.service = self._service
        for key, value in get_log_extras().items():
            setattr(record, key, value)
        return True


def uvicorn_server_factory(parser: ArgumentParser):
    from fastapi import FastAPI
    argument_group = parser.add_argument_group(title='Uvicorn server arguments')
    argument_group.add_argument('-remote', action='store_true', help='should listen for remote connections')
    argument_group.add_argument('-port', type=int, help='port to listen on', default=8000)
    argument_group.add_argument('-logging_conf', help='path to json file with logging config',
                                default='talisman-tools/default_logging_conf.json')

    def launcher(app: FastAPI, args: Namespace) -> None:
        import logging.config
        import logging.handlers
        import uvicorn
        from talisman_tools.configure.configure import read_config

        logging.config.dictConfig(read_config(args.logging_conf))

        f = TalismanFilter()
        for handler in logging.root.handlers:
            handler.addFilter(f)

        host = "0.0.0.0" if args.remote else "127.0.0.1"
        uvicorn.run(app, host=host, port=args.port, log_config=None)

    return launcher
